#!/usr/bin/env python
"""
synopsis:
    Search multiple XML documents using xpath.
    Given an xpath query, search multiple XML files with that query.
    Report back information on results.
usage:
    python searchxml.py [options] file1 file2 ...
examples:
    python searchxml.py ".//anelement" *.xml
    python searchxml.py -f -o results.log ".//anelement" *.xml
    python searchxml.py -v "names lines" ".//anelement" *.xml
    python searchxml.py -v "progress names lines" ".//anelement" *.xml
notes:
    For the default namespace, use prefix "abc:".  E.g. for elements in
    namespace xmlns="http://x.y.z", use query ".//abc:tag".
"""


import sys
import os
import argparse
import glob
from lxml import etree


def search(args):
    """Search each file.

    :param args:

    """
    filenames = generate_filenames(args)
    filenames.sort()
    for filename in filenames:
        search_file(filename, args)
    if args.countmatches is not None:
        args.outfile.write("--------------------\n")
        args.outfile.write(f"total matches: {args.countmatches}\n")


def search_file(filename, args):
    """Search one file and display the results.

    :param filename:
    :param args:

    """
    if "progress" in args.verbose:
        args.outfile.write("searching doc: {}\n".format(filename))
    doc = etree.parse(filename)
    root = doc.getroot()
    nsmap = root.nsmap
    if None in nsmap:
        nsmap["abc"] = nsmap[None]
        nsmap.pop(None)
    else:
        nsmap["abc"] = "http://nowhere.and.nothing.org"
    for key, value in args.extransprefixes.iteritems():
        if key not in nsmap:
            nsmap[key] = value
    results = root.xpath(args.query, namespaces=nsmap)
    if args.countmatches is not None:
        args.countmatches += len(results)
    display_results(results, filename, args)


def display_results(results, filename, args):
    """Display the results from one query.
    Respect command line options for how much and which info to display.

    :param results:
    :param filename:
    :param args:

    """
    if "progress" in args.verbose:
        fill = "    "
    else:
        fill = ""
    for result in results:
        if isinstance(result, etree._Element):
            if "names" in args.verbose:
                if "lines" in args.verbose:
                    args.outfile.write(
                        "{}{}:{}: {}\n".format(
                            fill, filename, result.sourceline, result.tag
                        )
                    )
                else:
                    args.outfile.write(
                        "{}{}: {}\n".format(fill, filename, result.tag))
            else:
                if "lines" in args.verbose:
                    args.outfile.write(
                        "{}{}: {}\n".format(
                            fill, result.sourceline, result.tag)
                    )
                else:
                    args.outfile.write("{}{}\n".format(fill, result.tag))
        else:
            if "names" in args.verbose:
                if "lines" in args.verbose:
                    if result.getparent() is None:
                        node = result
                    else:
                        node = result.getparent()
                    args.outfile.write(
                        "{}{}:{}: {}\n".format(
                            fill, filename, node.sourceline, result)
                    )
                else:
                    args.outfile.write("{}{}: {}\n".format(
                        fill, filename, result))
            else:
                if "lines" in args.verbose:
                    if result.getparent() is None:
                        node = result
                    else:
                        node = result.getparent()
                    args.outfile.write(
                        "{}{}: {}\n".format(fill, node.sourceline, result)
                    )
                else:
                    args.outfile.write("{}{}\n".format(fill, result))


def generate_filenames(args):
    """Generate and return a list of file names.

    :param args:
    :returns: if requested.

    """
    if args.recursive:
        filenames = []
        for directory, _, _ in os.walk(args.pathroot):
            filenames.extend(
                glob.glob(os.path.join(directory, args.filenamespec)))
    else:
        filenames = glob.glob(args.filenamespec)
    return filenames


def open_file(args):
    """

    :param args:

    """
    if not args.force:
        if os.path.exists(args.outfile):
            sys.exit("\noutput file exists.  use -f/--force to overwrite.\n")
    outfile = open(args.outfile, "w")
    return outfile


def main():
    """ main function """
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "query", type=str, help="xpath query.  applied to root node.")
    parser.add_argument(
        "filenamespec",
        metavar="filespec",
        type=str,
        help="file name spec/pattern for files to be searched",
    )
    parser.add_argument(
        "-p",
        "--pathroot",
        dest="pathroot",
        type=str,
        action="store",
        default=".",
        metavar="pathroot",
        help="path to directory for search or (if recursive) "
        "start of search.  Default: '.'",
    )
    parser.add_argument(
        "-r",
        "--recursive",
        dest="recursive",
        action="store_true",
        default=False,
        help="also search subdirectories.  default: False",
    )
    parser.add_argument(
        "-o",
        "--outfile",
        dest="outfile",
        type=str,
        action="store",
        default="",
        metavar="outfilename",
        help="file to write results to.  default: sys.stdout",
    )
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=False,
        help="force overwrite of output file.",
    )
    parser.add_argument(
        "-x",
        "--extransprefixes",
        dest="extransprefixes",
        action="store",
        type=str,
        default="",
        metavar="extransprefixes",
        help="extra namespace prefixes to be given dummy definitions.",
    )
    parser.add_argument(
        "-c",
        "--countmatches",
        dest="countmatches",
        action="store_true",
        default=False,
        help="also write total number of matches.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store",
        type=str,
        default="",
        metavar="verbose",
        help="print additional info and actions"
        "progress', 'names' (file names), 'lines (line numbers)."
        "default: ''",
    )
    args = parser.parse_args()
    if args.outfile:
        args.outfile = open_file(args)
    else:
        args.outfile = sys.stdout
    prefixdict = {}
    if args.extransprefixes:
        prefixes = args.extransprefixes.split()
        for prefix in prefixes:
            prefixdict[prefix] = "http://www.nowhereandnoplace.org"
    args.extransprefixes = prefixdict
    args.verbose = set(args.verbose.split())
    if args.countmatches:
        args.countmatches = 0
    else:
        args.countmatches = None
    search(args)


if __name__ == "__main__":
    # import ipdb; ipdb.set_trace()
    main()
